import asyncio
from typing import Any
from cowait.network import ON_CLOSE
from cowait.utils import Version
from ..status import WAIT, WORK, FAIL
from ..messages import TASK_INIT, TASK_STATUS, TASK_FAIL, TASK_RETURN
from ..instance import TaskInstance
from ..remote_task import RemoteTask


class TaskManager(dict):
    """ Keeps track of the state of subtasks """

    def __init__(self, task):
        self.task = task
        self.conns = {}

        # subscribe to child task status updates
        # maybe move it out?
        task.node.server.on(TASK_INIT, self.on_child_init)
        task.node.server.on(TASK_STATUS, self.on_child_status)
        task.node.server.on(TASK_RETURN, self.on_child_return)
        task.node.server.on(TASK_FAIL, self.on_child_fail)
        task.node.server.on(ON_CLOSE, self.on_child_close)

        # forward child events to parent
        # this probably belongs somewhere else...
        async def forward(conn, **msg):
            if msg.get('type', '__')[:2] == '__':
                return
            await task.node.parent.send(msg)
        task.node.server.on('*', forward)

    def watch(self, task, timeout=30):
        """
        Watch for a newly created task. Registers it with the task manager and sets up an
        initialization timeout check.
        """
        self[task.id] = task
        self.task.node.io.create_task(self._init_timeout_check(task, timeout))

    async def _init_timeout_check(self, task, timeout):
        await task.wait_for_scheduling()
        await asyncio.sleep(timeout)
        if not task.future.done() and task.id not in self.conns.values():
            # make sure this task is killed and cant connect after the failure
            # note: there is a race condition here.
            # the kill command might arrive after the error has been sent
            # perhaps we should note that this task has failed, and ignore any
            # further communication from it?
            self.task.cluster.destroy(task.id)

            # emit a timeout error
            await self.emit_child_error(
                id=task.id, error=f'{task.id} timed out before initialization')

    async def on_child_init(self, conn, id: str, task: dict, version: str, **msg: dict):
        # version check
        try:
            version = Version.parse(version)
            if not version.is_compatible():
                raise ValueError(f'{id} runs an incompatible library version: {version}')

        except ValueError as e:
            # illegal version string or incompatible version
            await self.reject(id, str(e), conn)
            return

        # register child task
        self.register(conn, id, task)

    async def reject(self, id, reason: str, conn=None):
        await self.emit_child_error(id=id, error=reason)
        if id in self:
            del self[id]
        if conn is not None:
            await conn.close()
            if conn in self.conns:
                del self.conns[conn]

    def register(self, conn, id, task: dict):
        # register connection if its not yet known
        if conn not in self.conns:
            self.conns[conn] = id

        if id in self:
            # we know about the task - this should mean that its a child of the current task.
            # if not, something weird is going on.
            task = self[id]
            task.conn = conn
        else:
            # this task is not yet known - its either a virtual task or a child of a child.
            # check if it has the running task set as its parent, if so, try to adopt it.
            taskdef = TaskInstance.deserialize(task)
            if taskdef.parent == self.task.id:
                # its a virtual task
                task = RemoteTask(taskdef, self.task.cluster)
                task.conn = conn
                self[id] = task

    async def on_child_status(self, conn, id: str, status: str, **msg: dict):
        if id in self:
            task = self[id]
            task.set_status(status)

    async def on_child_return(self, conn, id: str, result: Any, result_type: any, **msg: dict):
        if id in self:
            task = self[id]
            task.set_result(result, result_type)

    async def on_child_fail(self, conn, id: str, error: str, **msg: dict):
        if id in self:
            task = self[id]
            task.set_error(error)

    async def on_child_close(self, conn, **msg: dict) -> None:
        if conn not in self.conns:
            return

        # lookup task & remove from active connections
        task_id = self.conns[conn]
        task = self[task_id]
        del self.conns[conn]

        # unset task connection reference
        task.conn = None

        # emit an error event if connection was lost before done/fail
        if task.status == WAIT or task.status == WORK:
            await self.emit_child_error(
                conn=conn, id=task.id, error=f'Lost connection to {task_id}')

    async def emit_child_error(self, id, error, conn=None):
        await self.task.node.server.emit(type=TASK_STATUS, id=id, status=FAIL, conn=conn)
        await self.task.node.server.emit(type=TASK_FAIL, id=id, conn=conn, error=error)



import os
import json
from datetime import datetime
from .logger import Logger


class JSONLogger(Logger):
    def __init__(self):
        # create a duplicate of stdout early on, so that output from the logger
        # won't be captured and cause a infinite recursion
        self.stdout = os.dup(1)

    def __del__(self):
        os.close(self.stdout)

    def log(self, type: str, msg: dict) -> None:
        msg['ts'] = datetime.now().isoformat()
        jsonstr = json.dumps(msg) + '\n'
        os.write(self.stdout, jsonstr.encode('utf-8'))



import os
import pytest
import asyncio
from alt_pytest_asyncio.plugin import AltPytestAsyncioPlugin
from cowait import Task


class PytestTask(Task):
    async def run(
        self,
        marks=None,
        verbose: bool = True,
        capture: bool = True,
    ):
        plugins = [
            # make sure pytest uses the existing event loop
            AltPytestAsyncioPlugin(loop=asyncio.get_event_loop()),
        ]

        # use cowait's bundled pytest config if none is provided in the project root
        config = None
        if not os.path.exists('pytest.ini'):
            config = '/var/cowait/pytest.ini'

        args = create_pytest_args(
            config=config,
            verbose=verbose,
            marks=marks,
            capture=capture,
        )

        print('Args:', ' '.join(args))

        # run tests
        code = pytest.main(args, plugins=plugins)

        # throw an exception if the tests failed
        if code != pytest.ExitCode.OK and \
           code != pytest.ExitCode.NO_TESTS_COLLECTED:
            raise RuntimeError('Tests failed')

        return True


def create_pytest_args(config=None, marks=None, verbose=True, capture=True):
    args = []

    if config:
        args += ['-c', config]

    # verbose output
    if verbose:
        args.append('-vv')

    # output capture settings
    args.append('--capture=' + ('fd' if capture else 'no'))

    # marks
    if marks and len(marks) > 0:
        args += ["-m", marks]

    return args



import asyncio
from cowait.network import Server, get_local_ip, WS_PATH
from cowait.utils.file_capture import FileCapturing
from .io_thread import IOThread
from .logger import Logger
from .parent_client import ParentClient
from .resource_monitor import ResourceMonitor


class WorkerNode(object):
    def __init__(self, id: str, upstream: str, port: int = 80, logger: Logger = None):
        super().__init__()
        self.id = id
        self.io = IOThread()
        self.io.start()
        self.upstream = upstream
        self.server = Server(port=port)
        self.parent = ParentClient(id, self.io, logger)

    async def connect(self) -> None:
        await self.parent.connect(self.upstream)

    def serve(self):
        self.io.create_task(self.server.serve())

    def get_url(self):
        local_ip = get_local_ip()
        return f'ws://{local_ip}:{self.server.port}/{WS_PATH}'

    async def close(self) -> None:
        async def _close():
            await asyncio.sleep(0.2)
            await self.server.close()
            await self.parent.close()

        await asyncio.sleep(0.2)
        self.io.create_task(_close())

    def capture_logs(self) -> FileCapturing:
        """ Sets up a stream capturing context, forwarding logs to the node """
        def logger(file):
            def callback(x):
                nonlocal file
                self.io.create_task(self.parent.send_log(file, x))
            return callback

        return FileCapturing(
            on_stdout=logger('stdout'),
            on_stderr=logger('stderr'),
        )

    def monitor_system(self, interval: float = 1.0):
        async def monitor_loop():
            monitor = ResourceMonitor()
            while True:
                await asyncio.sleep(interval)
                try:
                    await self.parent.send_stats(monitor.stats())
                except ConnectionResetError:
                    return

        self.io.create_task(monitor_loop())



import os
import io
import sys
from tempfile import TemporaryFile

STDOUT = 1
STDERR = 2
SYS_FDS = {
    STDOUT: 'stdout',
    STDERR: 'stderr',
}


class CallbackFile(io.TextIOWrapper):
    def __init__(self, callback: callable = None):
        super().__init__(
            TemporaryFile(buffering=0),
            encoding='utf-8',
            errors='replace',
            write_through=True,
            newline='',
        )
        self.callback = callback

    def write(self, data) -> None:
        super().write(data)
        if '\n' in data:
            self.flush()

    def getvalue(self) -> str:
        self.buffer.seek(0)
        res = self.buffer.read()
        self.buffer.seek(0)
        self.buffer.truncate()
        return res.decode('utf-8')

    def flush(self) -> None:
        if self.callback is not None:
            text = self.getvalue()
            if len(text) > 0:
                self.callback(text)
        else:
            super().flush()


class FDCapture:
    def __init__(self, targetfd: int, callback: callable = None) -> None:
        self.targetfd = targetfd
        self.targetfd_save = os.dup(targetfd)
        self.callback = callback
        self.tmpfile = CallbackFile(callback)
        self._capturing = False

    def start(self) -> None:
        """Start capturing on targetfd using memorized tmpfile."""
        assert not self._capturing

        # redirect the target fd to our temporary file
        os.dup2(self.tmpfile.fileno(), self.targetfd)

        if self.targetfd in SYS_FDS:
            # for system file descriptors (stdout, stderr) we also need to replace the sys.stdout/sys.stderr streams.
            # this allows us to capture output from the local python process
            self.sysfile = getattr(sys, SYS_FDS[self.targetfd])
            setattr(sys, SYS_FDS[self.targetfd], self.tmpfile)

        self._capturing = True

    def stop(self) -> None:
        """Stop capturing, restore streams & return original capture file"""
        assert self._capturing

        # disable output redirection
        os.dup2(self.targetfd_save, self.targetfd)

        # if we have a callback, flush any remaining data
        if self.callback:
            remainder = self.getvalue()
            if len(remainder) > 0:
                self.callback(remainder)

        # close temporary file descriptor
        os.close(self.targetfd_save)

        if self.targetfd in SYS_FDS:
            # restore the original sys stream
            setattr(sys, SYS_FDS[self.targetfd], self.sysfile)

        self._capturing = False

    def writeorg(self, data) -> None:
        """Write to original file descriptor."""
        if self._capturing:
            os.write(self.targetfd_save, data.encode('utf-8'))
        else:
            os.write(self.targetfd, data.encode('utf-8'))

    def getvalue(self) -> str:
        return self.tmpfile.getvalue()

    def __del__(self) -> None:
        self.tmpfile.close()


class FileCapturing(object):
    def __init__(
        self,
        on_stdout: callable = None,
        on_stderr: callable = None,
    ):
        self.stdout = FDCapture(STDOUT, on_stdout)
        self.stderr = FDCapture(STDERR, on_stderr)

    def __enter__(self) -> None:
        self.stdout.start()
        self.stderr.start()
        return self

    def __exit__(self, *args) -> None:
        self.stdout.stop()
        self.stderr.stop()

    def __del__(self) -> None:
        del self.stdout
        del self.stderr
